#include "main.h"
#include "displayManager.h"
#include "Button.h"
#include <TJpg_Decoder.h>
#include "icons.h"
#include <WiFi.h>
#include <HTTPClient.h>
#include <PubSubClient.h>
#include <time.h>

TFT_eSPI tft = TFT_eSPI();

Button myButtonList[NUMBEROFDISPLAYS];

struct myEntry 
{
  boolean hasConnectionTasmota = false;
  boolean hasStartMQTT = false;
  boolean hasPowerDownMQTT = false;
  int16_t myActWattTasmota = -1;
  boolean hasPowerDownMQTTScript = true;
  String myPowerDownMQTTTopic = "";
  String myPowerDownMQTTLoad = "";
  uint8_t myStandbyWattTasmota = 6;
  String myIPTasmota = "";
  String mySwitchOffTasmota = "";
  String mySwitchOnTasmota = "";
  uint64_t lastCheckedTimestamp = 0l;
  uint64_t lastActiveTimestamp = 0l;
  int8_t myLastDisplayedonTFT = DISPLAY_UNDEFINED;
  boolean myCanDisplayClock = true;
  boolean myShowsInfoMessage = false;
  uint64_t myInfoMessageTimeStamp = 0l;
  uint64_t myInfoMessageDuration = MYINFODISPLAYTIMESECONDS * 1000;
  boolean myShowsAlarmMessage = false;
  boolean myShowsAlarmText = false;
  boolean myShowsAlarmIcon = false;
  boolean myIsTrueNAS = false;
  String myTrueNASAPIKey = "";
  String myTrueNASAPIUrl = "";
};


myEntry  myEntryList[NUMBEROFDISPLAYS];


WiFiClient myESPClient;
PubSubClient myMQTTClient(myESPClient); 

displayManager* myDisplayManager;


//--------------------------------------------------------------------------------------------------------------------------------------------

bool tft_output(int16_t x, int16_t y, uint16_t w, uint16_t h, uint16_t *myBitmap) 
{
  TFT_eSPI myDisplay = myDisplayManager->getDisplay();

  if (y >= myDisplay.height())
    return false;
  myDisplay.pushImage(x, y, w, h, myBitmap);
  return true;
}

//------------------------------------------- Uhr Ziffenblatt ----------------------------------------------------------------------------------------------

void AnalogClock_createFace (int8_t myIndex)
{
  uint16_t i;
  uint16_t myRadius;

  // Dann sollte das bereits da sein
  if (myEntryList[myIndex].myLastDisplayedonTFT == DISPLAY_ANALOGCLOCK)
    return;
  
  if (myEntryList[myIndex].myShowsAlarmMessage) 
    return;

  myDisplayManager->selectScreen(myIndex);
  TFT_eSPI myDisplay = myDisplayManager->getDisplay();

  myRadius = myDisplay.width()/2;

  myDisplay.fillCircle(myRadius, myRadius, myRadius, TFT_DARKCYAN);
  myDisplay.fillCircle(myRadius, myRadius, (myRadius-3), TFT_BLUE);
  myDisplay.fillCircle(myRadius, myRadius, (myRadius-10), TFT_BACKGROUND_COLOR);   

  // Stundenmarker
  for (i = 0; i < 360; i += 30)                                                     
  {                                                   
    sx = cos((i-90) * DEG2RAD);
    sy = sin((i-90) * DEG2RAD);
    x0 = sx * (myRadius-6) + myRadius;
    yy0 = sy * (myRadius-6) + myRadius;
    x1 = sx * (myRadius-20) + myRadius;
    yy1 = sy * (myRadius-20) + myRadius;
    myDisplay.drawLine(x0, yy0, x1, yy1, TFT_ORANGE);
  }
                                                             
  // Minuten Punkte
  for (i = 0; i < 360; i += 6)                                                      
  {
    sx = cos((i-90) * DEG2RAD);
    sy = sin((i-90) * DEG2RAD);
    x0 = sx * (myRadius-18) + myRadius;
    yy0 = sy * (myRadius-18) + myRadius;    
    myDisplay.drawPixel(x0, yy0, TFT_WHITE);
    
    if(i == 0  || i == 180) myDisplay.fillCircle (x0, yy0, 2, TFT_WHITE);                         
    if(i == 90 || i == 270) myDisplay.fillCircle (x0, yy0, 2, TFT_WHITE);
  }
  
  myDisplay.fillCircle(myRadius, (myRadius+1), 3, TFT_WHITE);                                     
  targetTime = millis() + 1000; 
  myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_ANALOGCLOCK;  
}

//-------------------------------------------------------------------------------------------------------------------------------------

void AnalogClock_updateZeiger(int8_t myIndex)
{
  int16_t myRadius = 0;
  
  if (targetTime < millis())
  {

    targetTime += 1000;

    myAnalogClockSecond++;                                                                            

    if (myAnalogClockSecond >= 60)
    {
      myAnalogClockSecond = 0;
      myAnalogClockMinute++;  
      if(myAnalogClockMinute > 59)
      {        
        myAnalogClockMinute=0;
        AnalogClock_createFace(myIndex);
        myAnalogClockHour++;
        if (myAnalogClockHour > 24)
        {
          myAnalogClockHour=0;
          myAnalogClockDateString = "  ";
        }
      }
      
      if ((myAnalogClockMinute % 5) == 0)
      {
        if (getInternetTime() == false)
        {
          myEntryList[myIndex].lastActiveTimestamp = millis();
          myDisplayManager->selectScreen(myIndex);
          TJpgDec.drawJpg(0, 0, powerbtn_off, sizeof(powerbtn_off));
          myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_POWERBTN;
          return;
        }
      }
    }

    if (myEntryList[myIndex].myLastDisplayedonTFT != DISPLAY_ANALOGCLOCK)
      return;

    myDisplayManager->selectScreen(myIndex);
    TFT_eSPI myDisplay = myDisplayManager->getDisplay();
    myRadius = myDisplay.width() / 2;

    sdeg = myAnalogClockSecond * 6;                                                                     
    mdeg = myAnalogClockMinute * 6 + sdeg * 0.01666667;                                                     
    hdeg = myAnalogClockHour * 30 + mdeg * 0.0833333;                                                     
    hx = cos ((hdeg-90) * DEG2RAD);    
    hy = sin ((hdeg-90) * DEG2RAD);
    mx = cos ((mdeg-90) * DEG2RAD);    
    my = sin ((mdeg-90) * DEG2RAD);
    sx = cos ((sdeg-90) * DEG2RAD);    
    sy = sin ((sdeg-90) * DEG2RAD);

    if (myAnalogClockSecond == 0 || firstStartAnalogClock) 
    {
      firstStartAnalogClock = false;
      myDisplay.drawLine (ohx, ohy, myRadius, (myRadius+1), TFT_BACKGROUND_COLOR);                                     
      ohx = hx * (myRadius - 58) + (myRadius + 1);    
      ohy = hy * (myRadius - 58) + (myRadius + 1);
      myDisplay.drawLine (omx, omy, myRadius, (myRadius+1), TFT_BACKGROUND_COLOR);
      omx = mx * 84 + myRadius;    
      omy = my * 84 + myRadius + 1;
    }
 
    myDisplay.drawLine (osx, osy, myRadius, (myRadius+1), TFT_BACKGROUND_COLOR);                                      
    osx = sx * (myRadius-30) + (myRadius+1);    
    osy = sy * (myRadius-30) + (myRadius+1);
    
    myDisplay.setTextFont(0);

    if (myAnalogClockDateString.compareTo(myAnalogClockOldDateString) != 0)
    {
      myDisplay.setTextColor(TFT_BACKGROUND_COLOR);
      myDisplay.drawString(myAnalogClockOldDateString, (myRadius+68), (myRadius-8), 4);
      myAnalogClockOldDateString = myAnalogClockDateString;
    }
    myDisplay.setTextColor(TFT_ORANGE);
    myDisplay.drawString(myAnalogClockDateString, (myRadius+68), (myRadius-8), 4);

    myDisplay.drawLine (osx, osy, myRadius, myRadius+1, TFT_RED);
    myDisplay.drawLine (ohx, ohy, myRadius, myRadius+1, TFT_WHITE);
    myDisplay.drawLine (omx, omy, myRadius, myRadius+1, TFT_WHITE);
    myDisplay.drawLine (osx, osy, myRadius, myRadius+1, TFT_RED);
    myDisplay.fillCircle(myRadius, myRadius+1, 3, TFT_RED);
  }
}

//-------------------------------------------------------------------------------------------------------------------------------------------------

boolean getInternetTime()
{
  if (!WiFi.isConnected())
    connectToWifi();

  struct tm myTimeInfo;

  if(getLocalTime(&myTimeInfo))
  {
    myAnalogClockHour = myTimeInfo.tm_hour;
    myAnalogClockMinute = myTimeInfo.tm_min;
    myAnalogClockSecond = myTimeInfo.tm_sec;

    if (myTimeInfo.tm_mday > 9)
         myAnalogClockDateString = String(myTimeInfo.tm_mday);
    else myAnalogClockDateString = " " + String(myTimeInfo.tm_mday);

    #ifdef DEBUGENABLED
      Serial.printf("getInternetTime() Uhrzeit: %i:%i.%i\n\r", myAnalogClockHour, myAnalogClockMinute, myAnalogClockSecond);
      Serial.printf("getInternetTime() Datum Tag: %s\n\r", myAnalogClockDateString.c_str());
    #endif
    return(true);
  }
  else
  {
    #ifdef DEBUGENABLED
      Serial.printf("getInternetTime() Fehler getLocalTime()\n\r");
    #endif
    myAnalogClockHour = 0;
    myAnalogClockMinute = 0;
    myAnalogClockSecond = 0;
    myAnalogClockDateString = "  ";
    return(false);
  }
}


//-------------------------------------------------------------------------------------------------------------------------------------------------

int8_t getFreeInfoScreen()
{
  int8_t i;
  for (i=0; i < NUMBEROFDISPLAYS; i++)
  {
    if ((myEntryList[i].myLastDisplayedonTFT == DISPLAY_UNDEFINED) || (myEntryList[i].myLastDisplayedonTFT == DISPLAY_POWERBTN) || (myEntryList[i].myLastDisplayedonTFT == DISPLAY_ANALOGCLOCK))
    {
      return(i);
    }
  }
  for (i=0; i < NUMBEROFDISPLAYS; i++)
  {
    if ((myEntryList[i].myLastDisplayedonTFT == DISPLAY_WATTAGE))
      return(i);
  }
  return (-1);
}

//-------------------------------------------------------------------------------------------------------------------------------------------------

String httpGETRequest(const char* myServerName) 
{
  WiFiClient myClient;
  HTTPClient myHTTPClient;
    
  myHTTPClient.begin(myClient, myServerName);
  
  int httpResponseCode = myHTTPClient.GET();

  String myPayload = "";
  
  if (httpResponseCode > 0) 
  {
    #ifdef DEBUGENABLED
      Serial.printf("httpGETRequest HTTP Response [%s] code: ", myServerName);
      Serial.println(httpResponseCode);
    #endif 
    myPayload = myHTTPClient.getString();
  }
  else 
  {
    #ifdef DEBUGENABLED
      Serial.printf("httpGETRequest HTTP Response [%s] ErrorCode: %i\n\r", myServerName, httpResponseCode);
    #endif 
  }

  myHTTPClient.end();

  return myPayload;
}

//-------------------------------------------------------------------------------------------------------------------------------------------------

void readTasmotaData(uint8_t index)
{
  if (index < NUMBEROFDISPLAYS)
  {
     if (myEntryList[index].hasConnectionTasmota)
     {
        
        myEntryList[index].myActWattTasmota = -1;
        String abc = httpGETRequest(myEntryList[index].myIPTasmota.c_str());

        #ifdef DEBUGENABLED
          Serial.printf("HTTP Request an %s\n\r", myEntryList[index].myIPTasmota.c_str());
          Serial.printf("Antwort: %s\n\r", abc.c_str());
        #endif
      
        if (abc.indexOf("\"Power\":") > 0)
        {
           abc = abc.substring(abc.indexOf("\"Power\":")+8);
           if (abc.indexOf(",") > 0)
           {
              abc = abc.substring(0, abc.indexOf(","));             
           }
           myEntryList[index].myActWattTasmota = (int16_t) abc.toInt();
        }
        else myEntryList[index].myActWattTasmota = -1;
     }
  }  
}


void showAlarm(String myMessage1, String myMessage2)
{
  uint8_t i;

  for (i=0; i < NUMBEROFDISPLAYS; i++)
  {
    myEntryList[i].myLastDisplayedonTFT = DISPLAY_ALARM;
    myEntryList[i].myShowsAlarmMessage = true;
    if (myEntryList[i].myShowsAlarmIcon)
    {
      myDisplayManager->selectScreen(i); 
      TFT_eSPI myDisplay = myDisplayManager->getDisplay();
      TJpgDec.drawJpg(0, 0, alarm_icon, sizeof(alarm_icon));
    }
    else
    {
      myDisplayManager->selectScreen(i);
      TFT_eSPI myDisplay = myDisplayManager->getDisplay();
      myDisplay.fillScreen(TFT_RED);
      myDisplay.fillCircle(120,120, 102, TFT_BLACK);

      myDisplay.setTextFont(0);  
      myDisplay.setTextColor(TFT_WHITE);
  
      myDisplay.drawCentreString(myMessage1.c_str(), 120, 80, 4);
      myDisplay.drawCentreString(myMessage2.c_str(), 120, 140, 4);
    }
  }
}

//---------------------------------------------------------------------------------------

void doMQTTPowerDown(int myIndex)
{
  if ((myEntryList[myIndex].myPowerDownMQTTTopic.length() > 0) && (myEntryList[myIndex].myPowerDownMQTTLoad.length() > 0))
  {
    #ifdef DEBUGENABLED
      Serial.printf("doMQTTPowerDown myIndex: %i\n\r", myIndex);
    #endif
  
    MQTTReconnect();
    myMQTTClient.publish(myEntryList[myIndex].myPowerDownMQTTTopic.c_str(), myEntryList[myIndex].myPowerDownMQTTLoad.c_str());
    #ifdef DEBUGENABLED
      Serial.printf("doMQTTPowerDown publish Topic: %s, Load:%s\n\r", myEntryList[myIndex].myPowerDownMQTTTopic.c_str(), myEntryList[myIndex].myPowerDownMQTTLoad.c_str());
    #endif
  }
}

//--------------------------------------------------------------------------------------

boolean doTasmotaSwitchOff(int myIndex)
{
  String myResult = httpGETRequest(myEntryList[myIndex].mySwitchOffTasmota.c_str());

  if (myResult == nullptr)
  {
    #ifdef DEBUGENABLED
      Serial.printf("doTasmotaSwitchOff  httpGETRequest hat NULL zurueckgeliefert! KRAZZ\n\r");
    #endif
    return false;
  }

  #ifdef DEBUGENABLED
    Serial.printf("doTasmotaSwitchOff[%i] Request: %s\n\r", myIndex, myEntryList[myIndex].mySwitchOffTasmota.c_str());
    Serial.printf("doTasmotaSwitchOff[%i] Result: %s\n\r", myIndex, myResult.c_str());
  #endif
  if (myResult.indexOf("{\"POWER\":\"OFF\"}") >= 0)
       return true;
  else return false;
}

//---------------------------------------------------------------------

boolean doTasmotaSwitchOn(int myIndex)
{
  String myResult = httpGETRequest(myEntryList[myIndex].mySwitchOnTasmota.c_str());

  if (myResult == nullptr)
  {
    #ifdef DEBUGENABLED
      Serial.printf("doTasmotaSwitchOn  httpGETRequest hat NULL zurueckgeliefert! KRAZZ\n\r");
    #endif
    return false;
  }

  #ifdef DEBUGENABLED
    Serial.printf("doTasmotaSwitchOn[%i] Request: %s\n\r", myIndex, myEntryList[myIndex].mySwitchOffTasmota.c_str());
    Serial.printf("doTasmotaSwitchOn[%i] Result: %s\n\r", myIndex, myResult.c_str());
  #endif
  if (myResult.indexOf("{\"POWER\":\"ON\"}") >= 0)
       return true;
  else return false;
}

//---------------------------------------------------------------------------------

void connectToWifi()
{
  uint16_t i;

  for (i = 0; i < NUMBEROFDISPLAYS; i++)
    myEntryList[i].myLastDisplayedonTFT = DISPLAY_WIFICONNECT;

  uint8_t myActiveDisplay = 0;

  String myWiFiProgressDisplayString = "#";
  myDisplayManager->clearAllScreens();
  myDisplayManager->fillAllScreens(TFT_BACKGROUND_COLOR);
  myDisplayManager->selectScreen(myActiveDisplay);
  myDisplayManager->getDisplay().setTextFont(0);        
  myDisplayManager->getDisplay().setTextColor(TFT_WHITE);
  myDisplayManager->getDisplay().drawCentreString("WIFI", 120, 90, 4);
  myDisplayManager->getDisplay().drawCentreString("Connect", 120, 130, 4);
  if (NUMBEROFDISPLAYS > 0)
       myActiveDisplay++;
  else delay(2000);  
  
  myDisplayManager->selectScreen(myActiveDisplay);
  myDisplayManager->getDisplay().drawCentreString("SSID", 120, 90, 4);
  myDisplayManager->getDisplay().drawCentreString(WIFI_SSID, 120, 130, 4);
  
  WiFi.setAutoConnect(true);
  WiFi.enableSTA(true);
  
  if (NUMBEROFDISPLAYS > 1)
       myActiveDisplay++;
  else delay(2000);

  myDisplayManager->selectScreen(myActiveDisplay);
  myDisplayManager->getDisplay().setTextColor(TFT_GREENYELLOW);
  
  WiFi.begin(WIFI_SSID, WIFI_PASS);

  i = 0;

  while(WiFi.status() != WL_CONNECTED)
  {
    myDisplayManager->getDisplay().drawCentreString(myWiFiProgressDisplayString, 120, 110, 4);
    myWiFiProgressDisplayString += "#";
    if (myWiFiProgressDisplayString.length() > 10)
    {
      myWiFiProgressDisplayString = "";
      myDisplayManager->clearScreen(myActiveDisplay);
    }
    delay(100);
    i++;
    if (i > 600)          // Nach 1 Minute ohne WiFi Verbindungsaufbau reboot
    {
      myDisplayManager->clearAllScreens();
      myDisplayManager->selectScreen(myActiveDisplay);
      myDisplayManager->getDisplay().setTextColor(TFT_RED);
      myDisplayManager->getDisplay().drawCentreString("Kein", 120, 80, 4);
      myDisplayManager->getDisplay().drawCentreString("WiFi Connect", 120, 120, 4);
      myDisplayManager->getDisplay().drawCentreString("Starte neu", 120, 160, 4);
      delay(3500);
      myDisplayManager->clearAllScreens();
      esp_restart();
    }
  }

  #ifdef NTP_SERVER
    configTime(0, 0, NTP_SERVER);
    setenv("TZ", NTP_TIMEZONE, 1);
    tzset();
  #endif
}

//---------------------------------------------------------------------------------------------------------------------------------------------

void MQTTcallback(char* myTopic, byte* myPlayload, uint32_t myPayloadLength) 
{
  String myMessageTotal = "";
  String myLine1 = "";
  String myLine2 = "";
  #ifdef DEBUGENABLED
    Serial.print("MQTT Callback: [");
  #endif
  for (int i=0; i < myPayloadLength; i++) 
  {
    myMessageTotal += String( (char)myPlayload[i]);
  }
  #ifdef DEBUGENABLED
    Serial.printf("%s]\n\r", myMessageTotal.c_str());
  #endif  

  if (myMessageTotal.indexOf("\"line1\": \"") > 0)
  {
      myLine1 = myMessageTotal.substring(myMessageTotal.indexOf("\"line1\": \"")+10);
      myLine1.remove(myLine1.indexOf("\""));
  }
  if (myMessageTotal.indexOf("\"line2\": \"") > 0)
  {
      myLine2 = myMessageTotal.substring(myMessageTotal.indexOf("\"line2\": \"")+10);
      myLine2.remove(myLine2.indexOf("\""));
  }
  #ifdef DEBUGENABLED
    Serial.printf("MQTT Payload [%s], Line1: [%s], Line2: [%s]\n\r", myMessageTotal.c_str(), myLine1.c_str(), myLine2.c_str());
  #endif

  showAlarm(myLine1, myLine2);
}

//---------------------------------------------------------------------------------------------------------------------------------

void doTrueNASPowerDown(int8_t myIndex)
{
  HTTPClient myClient;

  String TempString = "";
  
  #ifdef DEBUGENABLED
    Serial.printf("Truenas Powerdown\n\r");
  #endif 

  myClient.begin(myEntryList[myIndex].myTrueNASAPIUrl);

  TempString = "Bearer " + myEntryList[myIndex].myTrueNASAPIKey;

  myClient.addHeader("Authorization", TempString, true, true);

  int httpResponseCode = myClient.POST("");

  #ifdef DEBUGENABLED
    Serial.printf("Truenas httpCode: %i\n\r", httpResponseCode);
  #endif
}


//----------------------------------------------------------------------------------------------------------------------------------------------------

void doSwitchOff(uint8_t myIndex)
{
  if (myIndex < NUMBEROFDISPLAYS)
  {
    if ((myEntryList[myIndex].myActWattTasmota > myEntryList[myIndex].myStandbyWattTasmota) && (myEntryList[myIndex].hasPowerDownMQTT))
    {
      doMQTTPowerDown(myIndex);
      return;
    }
    else if ((myEntryList[myIndex].myActWattTasmota > myEntryList[myIndex].myStandbyWattTasmota) && (myEntryList[myIndex].myIsTrueNAS))
    {
      doTrueNASPowerDown(myIndex);
      return;
    }
    else doTasmotaSwitchOff(myIndex);
  }
}

//--------------------------------------------------------------------------------------------------------------------------------------------------------------

void MQTTReconnect()
{
  if (myMQTTClient.connected())
    return;

  int myInfoScreen = getFreeInfoScreen();
  if (myInfoScreen >= 0)
    showInfoMessage(myInfoScreen, "MQTT Reconnect");

  while (!myMQTTClient.connected())
  {
    myMQTTClient.setServer(MQTTSERVER, MQTTSERVERPORT);  
    myMQTTClient.setCallback(MQTTcallback);
    myMQTTClient.connect(MQTTCLIENTID, MQTTUSER, MQTTPASSWORD);  
    #ifdef DEBUGENABLED
      Serial.println("MQTT (Re)connect....");
    #endif
    delay(250);
    if (!myMQTTClient.connected()) delay(2500);
  }
  String myString = String(MQTTCLIENTID) + "/AlarmInfo";
  myMQTTClient.publish(myString.c_str(), "Hallo");
  myString = String(MQTTCLIENTID) + "/AlarmInfo";
  #ifdef DEBUGENABLED
    Serial.printf("MQTT Connection: %i\n\r", myMQTTClient.connected());
    Serial.printf("MQTT Topic subscribed: %s\n\r", myString.c_str());
  #endif
  myMQTTClient.subscribe(myString.c_str());
}

//--------------------------------------------------------------------------------------------------------------------------------

void setup() 
{
  setCpuFrequencyMhz(CPU_FREQUENCY);


  #ifdef DEBUGENABLED
    Serial.begin(115200);
    Serial.println();
    Serial.println("Starte....");
  #endif


  for (int i=0; i < NUMBEROFDISPLAYS; i++)
  {
    if (i==0) myButtonList[0].beginWithPin(BUTTON_1);
    if (i==1) myButtonList[1].beginWithPin(BUTTON_2);
    if (i==2) myButtonList[2].beginWithPin(BUTTON_3);
  }


  myDisplayManager = new displayManager(tft);
  myDisplayManager->selectAllScreens();
  myDisplayManager->getDisplay().fillScreen(TFT_BLACK);
  myDisplayManager->reset();

  TJpgDec.setSwapBytes(true); // jpeg rendering setup
  TJpgDec.setCallback(tft_output);

  connectToWifi();
  MQTTReconnect();

  myDisplayManager->clearAllScreens();
  myDisplayManager->selectAllScreens();
  TJpgDec.drawJpg(0, 0, powerbtn_off, sizeof(powerbtn_off));

  for (uint8_t i = 0; i < NUMBEROFDISPLAYS; i++)
    myEntryList[i].myLastDisplayedonTFT = DISPLAY_POWERBTN;

  myEntryList[0].myIPTasmota = "http://192.168.0.53/cm?cmnd=status%2010";
  myEntryList[0].mySwitchOffTasmota = "http://192.168.0.53/cm?cmnd=Power%20Off";
  myEntryList[0].mySwitchOnTasmota = "http://192.168.0.53/cm?cmnd=Power%20On";
  myEntryList[0].hasConnectionTasmota = true;
  myEntryList[0].hasPowerDownMQTT = true;
  myEntryList[0].myPowerDownMQTTTopic = "homeassistant/button/FANLESS-8700G-satellite/FANLESS-8700G-satellite_shutdown/set";
  myEntryList[0].myPowerDownMQTTLoad = "PRESS";
  myEntryList[0].myShowsAlarmIcon = true;


  myEntryList[1].myIPTasmota = "http://192.168.0.60/cm?cmnd=status%2010";
  myEntryList[1].mySwitchOffTasmota = "http://192.168.0.60/cm?cmnd=Power%20Off";
  myEntryList[1].mySwitchOnTasmota = "http://192.168.0.60/cm?cmnd=Power%20On";
  myEntryList[1].hasConnectionTasmota = true;
  myEntryList[1].hasPowerDownMQTT = true;
  myEntryList[1].myPowerDownMQTTTopic = "homeassistant/button/-satellite_shutdown/set";
  myEntryList[1].myPowerDownMQTTLoad = "PRESS";
  myEntryList[1].myCanDisplayClock = true;

  myEntryList[2].myShowsAlarmIcon = true;
  myEntryList[2].myIsTrueNAS = true;
  myEntryList[2].myTrueNASAPIKey = "2-GLairuBQXVyBhwktqSndwZe1Xjgus5nfkoaNqWu8KCPG54mdW2FqokMkZKfq2rb9";
  myEntryList[2].myTrueNASAPIUrl = "http://192.168.0.8/api/v2.0/system/shutdown";
  myEntryList[2].hasStartMQTT = false;
  myEntryList[2].hasPowerDownMQTT = false;
  myEntryList[2].hasConnectionTasmota = true;
  myEntryList[2].myIPTasmota = "http://192.168.0.58/cm?cmnd=status%2010";
  myEntryList[2].mySwitchOffTasmota = "http://192.168.0.58/cm?cmnd=Power%20Off";
  myEntryList[2].mySwitchOnTasmota = "http://192.168.0.58/cm?cmnd=Power%20On";
  myEntryList[2].hasConnectionTasmota = true;

  for (int i=NUMBEROFDISPLAYS-1; i >= 0; i--)
    myEntryList[i].lastActiveTimestamp = millis();

}



void updateDisplay(uint8_t myIndex)
{
  if (myIndex >= NUMBEROFDISPLAYS) 
    return;

  if (myEntryList[myIndex].myShowsInfoMessage)
    return;

  if (myEntryList[myIndex].myShowsAlarmMessage)
    return;

  myDisplayManager->selectScreen(myIndex);
  TFT_eSPI myDisplay = myDisplayManager->getDisplay();

  if (myEntryList[myIndex].myActWattTasmota > 0)
  {
    if (myEntryList[myIndex].myLastDisplayedonTFT != DISPLAY_WATTAGE)
    {
      myDisplay.fillScreen(TFT_DARKCYAN);
      myDisplay.setTextColor(TFT_BLACK);
      myDisplay.setTextFont(1);
      if ((myEntryList[myIndex].hasPowerDownMQTT) && 
          (myEntryList[myIndex].myActWattTasmota > myEntryList[myIndex].myStandbyWattTasmota))
           myDisplay.drawCentreString("MQTT OFF", 120, 188, 4);
      else myDisplay.drawCentreString("Power OFF", 120, 188, 4);
      myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_WATTAGE;
    }
    myDisplay.fillRect(0,36,240,140, TFT_BLACK);
    myDisplay.setTextFont(0);
    myDisplay.setTextColor(TFT_WHITE);
    String abc = String(myEntryList[myIndex].myActWattTasmota);
    myDisplay.drawCentreString(abc.c_str(), 120, 55, 8);
    myDisplay.drawCentreString("Watt", 120, 140, 4);
    myEntryList[myIndex].lastActiveTimestamp = millis();
  }
  else if (myEntryList[myIndex].myActWattTasmota <= 0)
  {
    if (myEntryList[myIndex].myLastDisplayedonTFT != DISPLAY_ANALOGCLOCK)
    {
      if (myEntryList[myIndex].myLastDisplayedonTFT != DISPLAY_POWERBTN)
      {
        TJpgDec.drawJpg(0, 0, powerbtn_off, sizeof(powerbtn_off));
        myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_POWERBTN;
        myEntryList[myIndex].lastActiveTimestamp = millis();        
      }
    } 
  }
}

//------------------------------------------------------------------------------------------------------------------------------------------

boolean isClockDisplayed()
{
  for (uint8_t i=0; i < NUMBEROFDISPLAYS; i++)
  {
    if (myEntryList[i].myLastDisplayedonTFT == DISPLAY_ANALOGCLOCK) 
      return true;
  }
  return false;
}

//------------------------------------------------------------------------------------------------------------------------------------------

void showInfoMessage(int8_t myIndex, String myMessage)
{
  if (myIndex >= NUMBEROFDISPLAYS)
    return;
  
  if (myEntryList[myIndex].myShowsAlarmMessage)
    return;

  if (myMessage == NULL)
    myMessage = "";

  myDisplayManager->selectScreen(myIndex);
  TFT_eSPI myDisplay = myDisplayManager->getDisplay();
  myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_INFO;
  
  myDisplay.fillScreen(TFT_GREENYELLOW);
  myDisplay.setTextColor(TFT_BLACK);
  myDisplay.setTextFont(1);
  myDisplay.drawCentreString("INFO", 120, 200, 4);
  myDisplay.drawCentreString("INFO", 120, 16, 4);

  myDisplay.fillRect(0,50,240,140, TFT_BLACK);
  myDisplay.setTextFont(0);
  myDisplay.setTextColor(TFT_WHITE);

  myDisplay.drawCentreString(myMessage.c_str(), 120, 95, 4);
  myDisplay.drawCentreString("Bitte warten...", 120, 135, 2);
  myEntryList[myIndex].myShowsInfoMessage = true;
  myEntryList[myIndex].myInfoMessageTimeStamp = millis();
}

//----------------------------------------------------------------------------------------------------------------------------------------------

void clearInfoMessages()
{
  for (int i=0; i < NUMBEROFDISPLAYS; i++)
  {
    if (myEntryList[i].myShowsInfoMessage)
    {
      if (millis()-myEntryList[i].myInfoMessageTimeStamp > myEntryList[i].myInfoMessageDuration)
      {
        myEntryList[i].myShowsInfoMessage = false;
        myEntryList[i].myLastDisplayedonTFT = DISPLAY_UNDEFINED;
      }
    }
  }
}


void trueNASShutdown()
{

}

//-------------------------------------------------------------------------------------------------------------------------------------------------------------

void loop() 
{
  int16_t i;
  boolean myResult = false;
  int8_t myIndex = 0;

  boolean myClockDisplayed = isClockDisplayed();

  if (!WiFi.isConnected())
    connectToWifi();

  if (myMQTTClient.connected())
       myMQTTClient.loop();
  else MQTTReconnect();


  if (myClockDisplayed)
  {
    for (i=0; i < NUMBEROFDISPLAYS; i++)
    {
      if ((myEntryList[i].myLastDisplayedonTFT == DISPLAY_ANALOGCLOCK) && (!myEntryList[i].myShowsInfoMessage))
        AnalogClock_updateZeiger(i);
    }
  }


  if (!myEntryList[0].myShowsAlarmMessage)
  {
    clearInfoMessages();

    for (i=0; i < NUMBEROFDISPLAYS; i++)
    {
      if (millis()-myEntryList[i].lastCheckedTimestamp > (MYTASMOTACHECKINTERVALLSECONDS * 1000) )
      {
        myEntryList[i].lastActiveTimestamp = myEntryList[i].myActWattTasmota;
        readTasmotaData(i);
        updateDisplay(i);
        myEntryList[i].lastCheckedTimestamp = millis();
    
        if (!myClockDisplayed)
        {
          if (myEntryList[i].myCanDisplayClock)
          {
            #ifdef DEBUGENABLED
              Serial.printf("Kein Uhr angezeigt Checke Index %i %i\n\r", i, myEntryList[i].lastActiveTimestamp);
            #endif
            if ((millis() - myEntryList[i].lastActiveTimestamp > (1000 * MYANALOGCLOCKWAITTIME)) && (!myEntryList[i].myShowsInfoMessage) && (!myEntryList[i].myShowsAlarmMessage))
            {
              if (getInternetTime())
              {
                AnalogClock_createFace (i);
                AnalogClock_updateZeiger(i);
                myClockDisplayed = true;
              } 
              else myEntryList[i].lastActiveTimestamp = millis();   // Uhrzeit Hole Fehler.....  probiere spaeter wieder
            }
          }
        }
      }
    }
  }


  for (int8_t myIndex=0; myIndex < NUMBEROFDISPLAYS; myIndex++)
  {
    if (myEntryList[myIndex].myActWattTasmota > 0)
      myEntryList[myIndex].lastActiveTimestamp = millis();
 
    if (myButtonList[myIndex].pressed())
    {
      myEntryList[myIndex].lastActiveTimestamp = millis();
      if (myButtonList[myIndex].hasBeenPressed == false)
      {
        myButtonList[myIndex].hasBeenPressed = true;
      }
      else    // Doppelklick  -   MQTT immer umgehen, direkt auf Tasmota
      {
        #ifdef DEBUGENABLED
          Serial.printf("Button %i - DoubleClick\n\r", myIndex);
        #endif

        myButtonList[myIndex].hasBeenPressed = false;

        if (myEntryList[myIndex].myLastDisplayedonTFT == DISPLAY_ALARM)
        {
          myDisplayManager->selectAllScreens();
          TJpgDec.drawJpg(0, 0, powerbtn_off, sizeof(powerbtn_off));
          for (int j=0; j < NUMBEROFDISPLAYS; j++)
          {
            myEntryList[j].myLastDisplayedonTFT = DISPLAY_POWERBTN;
            myEntryList[j].myShowsAlarmMessage = false;
          }
          return;
        }

        TJpgDec.drawJpg(0, 0, powerbtn_pressed, sizeof(powerbtn_pressed));
        myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_POWERACTIVEBTN;

        if (myEntryList[myIndex].hasConnectionTasmota)
        {
          if (myEntryList[myIndex].myActWattTasmota == 0)
          {
            doTasmotaSwitchOn(myIndex);
            showInfoMessage(myIndex, "Tasmota ON");
          }
          else 
          {
            doTasmotaSwitchOff(myIndex);
            showInfoMessage(myIndex, "Tasmota OFF");
          }
        }
      }
    }

    if (myButtonList[myIndex].hasBeenPressed && (myButtonList[myIndex].getlastAction() > BUTTON_DOUBLECLICK_TIME))
    {
      myDisplayManager->selectScreen(myIndex);

      #ifdef DEBUGENABLED
        Serial.printf("Button %i - SinglePress\n\r", myIndex);
      #endif

      if (myEntryList[myIndex].myLastDisplayedonTFT == DISPLAY_ALARM)
      {
        myDisplayManager->selectAllScreens();
        TJpgDec.drawJpg(0, 0, powerbtn_off, sizeof(powerbtn_off));
        for (int j=0; j < NUMBEROFDISPLAYS; j++)
        {  
          myEntryList[j].myLastDisplayedonTFT = DISPLAY_POWERBTN;
          myEntryList[j].myShowsAlarmMessage = false;
        }
        myButtonList[myIndex].hasBeenPressed = false;
        return;
      }

      TJpgDec.drawJpg(0, 0, powerbtn_pressed, sizeof(powerbtn_pressed));
      myEntryList[myIndex].myLastDisplayedonTFT = DISPLAY_POWERACTIVEBTN;

      if (myEntryList[myIndex].myActWattTasmota > 0)
      {
        doSwitchOff(myIndex);
        showInfoMessage(myIndex, "Power OFF");
      }
      else if (myEntryList[myIndex].myActWattTasmota == 0)
      {
        doTasmotaSwitchOn(myIndex);
        showInfoMessage(myIndex, "Power ON");
      }

      updateDisplay(myIndex);

      myButtonList[myIndex].hasBeenPressed = false;
    }
  }

  //------------------------------------------------------------------------------------------------------------------------------


}